
import json
import csv
import logging
import matplotlib.pyplot as plt
from collections import defaultdict
import openpyxl

logging.basicConfig(level=logging.INFO)


class YotracoStats:
    def __init__(self):
        self.class_counts_in = defaultdict(int)
        self.class_counts_out = defaultdict(int)
    
    def get_counts(self):
        counts = {"in_counts" : dict(self.class_counts_in), "out_counts" : dict(self.class_counts_out)}
        logging.info("Fetched counts: %s", counts)
        return counts
    
    def save_counts(self, filename, file_format = "json"):
        data = self.get_counts()
        try :
            if file_format == 'json':
                with open(filename, "w") as f:
                    json.dump(data, f, indent=4)
                logging.info("Saved counts to JSON file: %s", filename)
            elif file_format == "csv":
                with open(filename, "w", newline='') as f:
                    writer = csv.writer(f)
                    writer.writerow(["Class", "IN Count", "OUT Count"])
                    for key in data["in_counts"].keys():
                        writer.writerow([key, data["in_counts"].get(key, 0), data["out_counts"].get(key,0)])
                logging.info("Saved counts to CSV file: %s", filename)
            elif file_format == "xlsx": # Supports Excel
                wb = openpyxl.Workbook()
                ws=wb.active
                ws.append(["Class" , "II Count", "OUT Count"])
                for key in data["in_counts"]:
                    ws.append([key,data["in_counts"].get(key,0),data["out_counts"].get(key,0)])
                wb.save(filename)
                logging.info("Saved counts to Excel file: %s", filename)
            elif file_format == "txt": # Supports TXT
                with open(filename,"w") as f:
                    f.write("Class\tIN Count\tOUT Count")
                    for key in data["in_counts"]:
                        f.write(f"{key}\t{data["in_counts"].get(key,0)}\t{data["out_counts"].get(key,0)}\n")
                logging.info("Saved counts to TXT file: %s", filename)
            else:
                # TODO : support other format
                logging.error("Unsupported file format: %s", file_format)
                raise ValueError("Unsupported file format. Use 'json', 'csv','xlsx' or 'txt' .")
        except Exception as e:
            logging.error("Error saving file %s: %s", filename, str(e))

    def plot_bar(self):
        try : 
            labels = list(set(self.class_counts_in.keys()).union(set(self.class_counts_out.keys())))
            in_counts = [self.class_counts_in.get(label, 0) for label in labels]
            out_counts = [self.class_counts_out.get(label, 0) for label in labels]
            if not labels:
                logging.warning("No data available for bar plot.")
                return
            x = range(len(labels))
            plt.figure(figsize=(10, 5))
            plt.bar(x, in_counts, width=0.4, label="IN", color="green" , align="center")
            plt.bar(x, out_counts, width=0.4, label="OUT", color="red", align="edge")
            plt.xticks(x, labels, rotation=45)
            plt.ylabel("Count")
            plt.title("Object Count Tracking")
            plt.legend()
            plt.show()
            logging.info("Displayed bar chart successfully.")
        except Exception as e:
            logging.error("Error generating bar chart: %s", str(e))

    def plot_pie(self):
        """ 
        This function displays two pie charts:
        - One for incominf objects (IN Count)
        - One for outgoing objects (OUT Count) 
        """
        try :
            labels_in = list(self.class_counts_in.keys())
            sizes_in = list(self.class_counts_in.values())
            labels_out = list(self.class_counts_out.keys())
            sizes_out = list(self.class_counts_out.values())

            if not labels_in and not labels_out:
                logging.warning("No data available for scatter plot.")
                raise ValueError("No data to display")
            fig, axs = plt.subplots(1, 2, figsize=(12,6))
            if labels_in:
                axs[0].pie(sizes_in, labels=labels_in, autopct="%1.1f%%",startangle=90,colors=plt.cm.Paired.colors)
                axs[0].set_title("Distribution of INCOMING objects")
            else:
                axs[0].set_title("No Incoming data") 

            if labels_out:
                axs[1].pie(sizes_out, labels=labels_out, autopct="%1.1f%%",startangle=90,colors=plt.cm.Paired.colors )
                axs[1].set_title("Distribution of Outgoing objects")
            else:
                axs[1].set_title("No Outgoing data ")
            plt.show()
            logging.info("Displayed pie chart successfully.")
        except Exception as e:
            logging.error("Error generating pie chart: %s", str(e))

    def plot_scatter(self):
        """ 
        This function displays a scatter plot of incoming VS outcoming object counts.
        """
        try :
            labels_in = list(self.class_counts_in.keys())
            labels_out = list(self.class_counts_out.keys())

            if not labels_in and not labels_out:
                logging.warning("No data available for scatter plot.")
                raise ValueError("No data to display")
            all_labels=list(set(labels_in).union(set(labels_out)))
            in_counts = [self.class_counts_in.get(label,0) for label in all_labels]
            out_count = [self.class_counts_out.get(label,0) for label in all_labels]
            
            plt.figure(figsize=(8, 6))
            plt.scatter(in_counts, out_count, color='green', label='Objects', alpha=0.7)
            plt.title("Scatter Plot of Incoming VS Outgoing Objects")
            plt.xlabel("Incoming Object Count")
            plt.ylabel("Outgoing Obejects Counts")
            plt.grid(True)
            plt.legend()
            plt.show()
            logging.info("Displayed scatter plot successfully.")
        except Exception as e:
            logging.error("Error generating scatter plot: %s", str(e))
        





import os
import cv2
import logging
# import threading
from ultralytics import YOLO
from collections import defaultdict
from YOTRACO.yotracoStats import YotracoStats

logging.basicConfig(level=logging.INFO)  # Set up logging globally

class Yotraco:

    def __init__(self, model_path, video_path, output_video
                #   output_format = 'mp4'
                  ,line_position='middle', track_direction='BOTH', 
                  classes_to_track=None , display=True):
        """
        Initialize the YOTRACO object with the specified YOLO model and video processing settings.

        Args:
        model_path (str): Path to the YOLO model file (e.g., a .pt or .onnx file).
        video_path (str): Path to the input video file to be processed.
        output_video (str): Path to save the processed video.
        line_position (str, optional): Vertical position of the tracking line ('top', 'middle', 'bottom'). Default is 'middle'.
        track_direction (str, optional): Direction to track ('BOTH', 'IN', or 'OUT'). Default is 'BOTH'.
        classes_to_track (list, optional): List of class indices to track. Default is [0, 1, 2, 3] (all classes).
        """

        self.stats = YotracoStats() 

        # Check if the model file exists
        if not os.path.exists(model_path):
            logging.info(f"Model file not found at'{model_path}' . \n Downloading...")
        else:
            logging.info(f"Model file found at '{model_path}'. \n Using it...")
            logging.info(f"Loading YOLO model from {model_path}...")

        self.model = YOLO(model_path)  # Load the model
        self.class_list = self.model.names  # Get class names

        logging.info("YOLO model loaded successfully.")

        # Open the video file
        self._cap = cv2.VideoCapture(video_path)
        if not self._cap.isOpened():
            raise ValueError("Error: Could not open video file.")

        self.display=display
        self.output_format = "avi"


        # Get video properties
        self.fps = self._cap.get(cv2.CAP_PROP_FPS)
        self.frame_width = int(self._cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.frame_height = int(self._cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        # Define output video settings based on format
        # fourcc_map = {
        #     'avi': cv2.VideoWriter_fourcc(*'XVID'),
        #     'mp4': cv2.VideoWriter_fourcc(*'mp4v'),
        #     'mov': cv2.VideoWriter_fourcc(*'MJPG')
        # }

        # if self.output_format not in fourcc_map:
        #     raise ValueError("Unsupported video format. Supported formats: avi, mp4, mov")

        # Define output video settings
        self.output_video = f"{output_video}.{self.output_format}"
        # TODO : support other extension
        # fourcc = fourcc_map[self.output_format]
        fourcc = cv2.VideoWriter_fourcc(*'MJPG')  # Codec for .avi format
        self.out = cv2.VideoWriter(self.output_video, fourcc, self.fps, (self.frame_width, self.frame_height))

        # TODO : add more control for the lines and add the abillity to put two line vertical and horizontal
        # Set line Y-coordinate based on position , the Y_line is dynamic for the horizantal line
        self.line_position = line_position

        #We need to fix the X_line in the middel for the vertical line
        # self.line_x = int(self._frame_height*0.5)

        # Initialize movement direction and classes to track
        self.track_direction = track_direction
        self.classes_to_track = classes_to_track if classes_to_track is not None else [0, 1, 2, 3]  # Default classes to track

        # Dictionaries for counting IN and OUT events
        self.class_counts_in = defaultdict(int)
        self.class_counts_out = defaultdict(int)
        self.crossed_ids = {}

    # track direction getter
    @property
    def track_direction(self):
        return self._track_direction
    
    # track direction setter
    @track_direction.setter
    def track_direction(self,direction):
        if direction in ['BOTH', 'IN', 'OUT']:
            self._track_direction = direction
        else :
            raise ValueError("track direction must be 'BOTH', 'IN', or 'OUT'")
    
    # line position getter 
    @property
    def line_position(self):
        return self._line_y
    
    # line position setter
    @line_position.setter
    def line_position(self,position):
        if position == 'top':
            self._line_y = int(self.frame_height * 0.3)
        elif position == 'bottom':
            self._line_y = int(self.frame_height * 0.7)
        elif position == 'middle':
            self._line_y = int(self.frame_height * 0.5)  
        else :
            raise ValueError("line position must be 'top', 'middle', 'bottom' ")
            

    def process_frame(self, frame):
        """
        Processes each frame to track objects using the YOLO model, detect crossing, and update counts.

        Args:
        frame (ndarray): A single frame from the video to process.
        """
        # Run YOLO tracking for specified classes
        results = self.model.track(frame, persist=True, classes=self.classes_to_track)

        # Draw tracking line across the full width of the frame
        cv2.line(frame, (0, self._line_y), (self.frame_width, self._line_y), (0, 0, 255), 3)

        # Process detections
        if results[0].boxes.data is not None:
            boxes = results[0].boxes.xyxy.cpu()
            if results[0].boxes.id is not None :
                track_ids = results[0].boxes.id.int().cpu().tolist()
            else:
                track_ids = []
            class_indices = results[0].boxes.cls.int().cpu().tolist()

            for box, track_id, class_idx in zip(boxes, track_ids, class_indices):
                x1, y1, x2, y2 = map(int, box)
                cx = (x1 + x2) // 2  # Calculate the center point of the bounding box
                cy = (y1 + y2) // 2
                class_name = self.class_list[class_idx]

                # Draw bounding box and tracking info
                cv2.circle(frame, (cx, cy), 4, (0, 0, 255), -1)
                cv2.putText(frame, f"ID: {track_id} {class_name}", (x1, y1 - 10),
                            cv2.FONT_HERSHEY_SIMPLEX, 0.6, (0, 255, 255), 2)
                cv2.rectangle(frame, (x1, y1), (x2, y2), (0, 255, 0), 2)

                # Track crossing direction
                if track_id not in self.crossed_ids:
                    self.crossed_ids[track_id] = cy
                else:
                    prev_cy = self.crossed_ids[track_id]
                    if self._track_direction == 'BOTH':  # Track both directions
                        if prev_cy < self._line_y <= cy:  # Moving downward (OUT)
                            self.class_counts_out[class_name] += 1
                            self.stats.class_counts_out[class_name] += 1
                        elif prev_cy > self._line_y >= cy:  # Moving upward (IN)
                            self.class_counts_in[class_name] += 1
                            self.stats.class_counts_in[class_name] += 1
                        self.crossed_ids[track_id] = cy
                    elif self._track_direction == 'IN' and prev_cy > self._line_y >= cy:  # Track only IN
                        self.class_counts_in[class_name] += 1
                        self.stats.class_counts_in[class_name] += 1
                        self.crossed_ids[track_id] = cy
                    elif self._track_direction == 'OUT' and prev_cy < self._line_y <= cy:  # Track only OUT
                        self.class_counts_out[class_name] += 1
                        self.stats.class_counts_out[class_name] += 1
                        self.crossed_ids[track_id] = cy
 
    def display_counts(self, frame):
        """
        Display the counts of objects that have crossed the line (both "IN" and "OUT").

        Args:
        frame (ndarray): A single frame from the video to overlay the count text on.
        """
        y_offset = 30
        for class_name, count in self.class_counts_out.items():
            cv2.putText(frame, f"OUT {class_name}: {count}", (50, y_offset),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.8, (0, 0, 255), 2)
            y_offset += 30

        y_offset = 30
        for class_name, count in self.class_counts_in.items():
            cv2.putText(frame, f"IN {class_name}: {count}", (self.frame_width - 250, y_offset),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.8, (0, 0, 255), 2)
            y_offset += 30

    def process_video(self):
        """
        Processes the video frame by frame, tracks objects, and saves the output video.

        Continuously processes the video, applies object detection and tracking, and saves the processed frames
        into the output video file.
        """
        while self._cap.isOpened():
            
            ret, frame = self._cap.read()
            if not ret:
                break
            
            self.process_frame(frame)
            if self.display==True:
                self.display_counts(frame)

            # Save processed frame
            self.out.write(frame)

        # Release resources after processing
        self._cap.release()
        self.out.release()


    # TODO : fix the multithreading 
    # def speed_process(self):
    #     threads = []
    #     while self._cap.isOpened():
    #         ret, frame = self._cap.read()
    #         if not ret:
    #             break

    #         # create a thread for processing each time
    #         thread = threading.Thread(target=self.process_frame, args=(frame,))
    #         threads.append(thread)
    #         thread.start()

    #         # limit the threads running at once 
    #         if len(threads)>5:
    #             for  t in threads:
    #                 t.join()
    #             threads = []


    #         # display the counts if display is true
    #         if self.display :
    #             self.display_counts(frame)

    #         self.out.write(frame) # save processed frame

    #     # wait for all threads to complete before releasing resources
    #     for t in threads:
    #         t.join()

    #     # release resources after processing
    #     self._cap.release()
    #     self.out.release()





import unittest
import json
import os
from YOTRACO.yotracoStats import YotracoStats
from unittest.mock import patch

class TestYotracoStats(unittest.TestCase):

    def setUp(self):
        """Set up a fresh stats object before each test."""
        self.stats = YotracoStats()
        self.test_json = "test_counts.json"
        self.test_csv = "test_counts.csv"

        # Ensure that counts contain at least one entry before testing save functions
        self.stats.counts = {"car": 5, "person": 10}  

    def tearDown(self):
        """Remove test files after each test."""
        for file in [self.test_json, self.test_csv]:
            if os.path.exists(file):
                os.remove(file)

    def test_save_counts_json(self):
        """Test saving the counts in JSON format."""
        self.stats.save_counts(self.test_json, "json")

        with open(self.test_json, "r") as f:
            data = json.load(f)

        self.assertTrue("in_counts" in data, "JSON should contain 'in_counts' key.")
        self.assertTrue("out_counts" in data, "JSON should contain 'out_counts' key.")

    # TODO : fix the csv test
    # def test_save_counts_csv(self):
    #     """Test saving the counts in CSV format."""
    #     self.stats.save_counts(self.test_csv, "csv")
        
    #     with open(self.test_csv, "r") as f:
    #         data = f.readlines()

    #     self.assertGreater(len(data), 1, "CSV should contain a header and at least one row.")

    @patch("matplotlib.pyplot.show")
    def test_plot_counts(self, mock_show):
        """Test plotting the counts (mocking `show` to avoid GUI issues)."""
        self.stats.plot_counts()
        mock_show.assert_called_once()  # Ensure the plot function is called

if __name__ == "__main__":
    unittest.main()



import unittest
from YOTRACO.yotraco import Yotraco
from PIL import Image
import numpy as np



class TestYotraco(unittest.TestCase):

    def setUp(self):
        """
        Set up mock objects and the environment before each test.
        """
        self.model_path = "tests/model/yolo11l.pt"
        self.video_path = "tests/videos/test_video.mp4"
        self.output_video = "output/video.avi"
        self.yotraco = Yotraco(self.model_path, self.video_path, self.output_video)

    def test_model_loading(self):
        """
        Test if the YOLO model is loaded correctly
        """
        self.assertTrue(self.yotraco.model is not None)
        # TODO : fix the model equality test
        # self.assertEqual(self.yotraco.model.model, self.model_path)


    def test_video_loading(self):
        """Test if the video is opened correctly"""
        self.assertTrue(self.yotraco.cap.isOpened())

    def test_process_frame(self):
        """Test the frame processing functionality"""
        frame = Image.new('RGB', (640, 480), color = (73, 109, 137))  # Create a real image
        frame = np.array(frame)
        self.yotraco.process_frame(frame)


if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages

setup(
    name="yotraco",
    version="0.2.4",  # Updated version
    packages=find_packages(include=["YOTRACO*"], exclude=["images"]),  # Match the inclusion/exclusion from pyproject
    install_requires=[
        "torch",
        "opencv-python",
        "ultralytics",
        "matplotlib",
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    author="YOTRACO TEAM",
    author_email="nereuscode@gmail.com",
    description="A YOLO-based object tracking and counting system with customizable tracking lines and movement direction.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords=['python', 'video', 'yolo', 'openCV'],
    url="https://github.com/NEREUS-code/YOTRACO",
    python_requires=">=3.8",
)
